/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/crypt/ofb64.c                                            *
 * Created:     2002-08-24 by Hampa Hug <hhug@student.ethz.ch>               *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "block64.h"
#include "ofb64.h"


void ofb_init (ofb_ctx_t *ctx, void *crypt, void *crypt_ctx)
{
	ctx->crypt = (ofb_crypt_f) crypt;
	ctx->ctx = crypt_ctx;

	ctx->buf_i = 0;
}

void ofb_set_iv_blk (ofb_ctx_t *ctx, const block64_t *iv)
{
	ctx->blk = *iv;

	blk64_set_buf (iv, ctx->buf);

	ctx->buf_i = 0;
}

void ofb_set_iv_buf (ofb_ctx_t *ctx, const unsigned char buf[8])
{
	blk64_get_buf (&ctx->blk, buf);
	blk64_set_buf (&ctx->blk, ctx->buf);

	ctx->buf_i = 0;
}

void ofb_crypt (ofb_ctx_t *ctx, unsigned char *buf, unsigned long n)
{
	unsigned long i;
	unsigned      key_i;
	unsigned char *key;

	key = ctx->buf;
	key_i = ctx->buf_i;

	for (i = 0; i < n; i++) {
		if (key_i == 0) {
			ctx->crypt (ctx->ctx, &ctx->blk, 1);
			blk64_set_buf (&ctx->blk, key);
		}

		buf[i] ^= key[key_i];

		key_i = (key_i + 1) & 7;
	}

	ctx->buf_i = key_i;
}
