/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/adler32.h                                            *
 * Created:     2006-06-39 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MISC_ADLER32_H
#define MISC_ADLER32_H 1


typedef struct {
	unsigned long s1;
	unsigned long s2;
} adler8_t;

typedef struct {
	unsigned long s1;
	unsigned long s2;
} adler16_t;

typedef struct {
	unsigned long s1;
	unsigned long s2;
} adler32_t;


void adler8_init (adler8_t *ad);
void adler8_copy (adler8_t *dst, const adler8_t *src);
void adler8_calc (adler8_t *ad, const void *buf, unsigned long cnt);
void adler8_done (adler8_t *ad);
unsigned adler8_get_sum (adler8_t *ad);

void adler16_init (adler16_t *ad);
void adler16_copy (adler16_t *dst, const adler16_t *src);
void adler16_calc (adler16_t *ad, const void *buf, unsigned long cnt);
void adler16_done (adler16_t *ad);
unsigned adler16_get_sum (adler16_t *ad);

void adler32_init (adler32_t *ad);
void adler32_copy (adler32_t *dst, const adler32_t *src);
void adler32_calc (adler32_t *ad, const void *buf, unsigned long cnt);
void adler32_done (adler32_t *ad);
unsigned long adler32_get_sum (adler32_t *ad);


#endif
