/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/sha256.h                                             *
 * Created:     2005-07-27 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MISC_SHA256_H
#define MISC_SHA256_H 1


/*!***************************************************************************
 * @short The SHA256 context
 *****************************************************************************/
typedef struct {
	unsigned      buf_i;
	unsigned char buf[64];

	unsigned long H[8];

	unsigned long cnt[2];
} sha256_t;


/*!***************************************************************************
 * @short Initialize an SHA256 context
 *****************************************************************************/
void sha256_init (sha256_t *sha);

/*!***************************************************************************
 * @short Copy an SHA256 context
 *****************************************************************************/
void sha256_copy (sha256_t *dst, const sha256_t *src);

/*!***************************************************************************
 * @short Calculate an SHA256 sum
 * @param sha The SHA256 context
 * @param buf The data
 * @param cnt The number of bytes at buf
 *
 * This function can be called multiple times to add more bytes.
 *****************************************************************************/
void sha256_calc (sha256_t *sha, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short End calculating an SHA256 sum
 * @param sha The SHA256 context
 *
 * This function must be called after the last call to sha256_calc() and
 * before getting the sha256 sum.
 *****************************************************************************/
void sha256_done (sha256_t *sha);

/*!***************************************************************************
 * @short Get a 32bit word of the SHA256 sum
 * @param sha The SHA256 context
 * @param i   The word index. 0 <= i <= 7.
 *****************************************************************************/
unsigned long sha256_get_sum_i (sha256_t *sha, unsigned i);

/*!***************************************************************************
 * @short Copy the SHA256 sum to a byte buffer
 * @param sha The SHA256 context
 * @param buf The 32 byte SHA256 buffer
 *****************************************************************************/
int sha256_get_sum (sha256_t *sha, void *buf, unsigned max);

/*!***************************************************************************
 * @short Calculate the SHA256 sum of a buffer
 * @param dst The 32 byte SHA256 buffer
 * @param src The source buffer
 * @param n   The size in bytes of the source buffer
 *****************************************************************************/
void sha256_calc1 (void *dst, const void *src, unsigned long n);


#endif
