/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/sha384.h                                             *
 * Created:     2005-11-18 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#ifndef MISC_SHA384_H
#define MISC_SHA384_H 1


/*!***************************************************************************
 * @short The SHA384 context
 *****************************************************************************/
typedef struct {
	unsigned           buf_i;
	unsigned char      buf[128];

	unsigned long long H[8];

	unsigned long long cnt[2];
} sha384_t;


/*!***************************************************************************
 * @short Initialize an SHA384 context
 *****************************************************************************/
void sha384_init (sha384_t *sha);

/*!***************************************************************************
 * @short Copy an SHA384 context
 *****************************************************************************/
void sha384_copy (sha384_t *dst, const sha384_t *src);

/*!***************************************************************************
 * @short Calculate an SHA384 sum
 * @param sha The SHA384 context
 * @param buf The data
 * @param cnt The number of bytes at buf
 *
 * This function can be called multiple times to add more bytes.
 *****************************************************************************/
void sha384_calc (sha384_t *sha, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short End calculating an SHA384 sum
 * @param sha The SHA384 context
 *
 * This function must be called after the last call to sha384_calc() and
 * before getting the sha384 sum.
 *****************************************************************************/
void sha384_done (sha384_t *sha);

/*!***************************************************************************
 * @short Get a 64bit word of the SHA384 sum
 * @param sha The SHA384 context
 * @param i   The word index. 0 <= i <= 5.
 *****************************************************************************/
unsigned long long sha384_get_sum_i (sha384_t *sha, unsigned i);

/*!***************************************************************************
 * @short  Copy the SHA384 sum to a byte buffer
 * @param  sha The SHA384 context
 * @param  buf The 48 byte SHA384 buffer
 * @param  max The size of the buffer
 * @return The number of bytes written to the buffer
 *****************************************************************************/
unsigned sha384_get_sum (sha384_t *sha, void *buf, unsigned max);

/*!***************************************************************************
 * @short Calculate the SHA384 sum of a buffer
 * @param dst The 48 byte SHA384 buffer
 * @param src The source buffer
 * @param n   The size in bytes of the source buffer
 *****************************************************************************/
void sha384_calc1 (void *dst, const void *src, unsigned long n);


#endif
