/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/subst/subst.c                                            *
 * Created:     2001-04-13 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/strarg.h>

#include "subst.h"


#define SUBST_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

static int        par_nobuf = 0;

static int        par_have_eof = 0;
static unsigned   par_eof_char = 0;
static int        par_eof = 0;

static unsigned   tab_byte[256];
static char       *tab_string[256];


static int tab_lcase[32] = {
	0x40, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
};

static int tab_ucase[32] = {
	0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
};

static int tab_nobin[128] = {
	 256,  256,  256,  256,  256,  256,  256,  256,
	 256, 0x09, 0x0a,  256,  256, 0x0d,  256,  256,
	 256,  256,  256,  256,  256,  256,  256,  256,
	 256,  256,  256,  256,  256,  256,  256,  256,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e,  256,
};

static int tab_ibmtoascii[256] = {
	 ' ',  '0',  '0',  '.',  '.',  '.',  '.',  '.', /* 00 */
	0x08, 0x09, 0x0a,  'o',  'o', 0x0d,  '.',  '.',
	 '>',  '<',  '|',  '!',  'P',  'P',  '_',  '|',
	 '^',  'v',  '>',  '<',  '-',  '-',  '^',  'v',
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e,  '^',
	 'C',  'u',  'e',  'a',  'a',  'a',  'a',  'c', /* 80 */
	 'e',  'e',  'e',  'i',  'i',  'i',  'A',  'A',
	 'E',  'a',  'f',  'o',  'o',  'o',  'u',  'u', /* 90 */
	 'y',  'O',  'U',  'c',  'L',  'Y',  'P',  'f',
	 'a',  'i',  'o',  'u',  'n',  'N',  'a',  'o', /* A0 */
	 '?',  '-',  '-',  '2',  '4',  '!',  '<',  '>',
	 '#',  '#',  '#',  '|',  '+',  '+',  '+',  '+', /* B0 */
	 '+',  '+',  '|',  '+',  '+',  '+',  '+',  '+',
	 '+',  '+',  '+',  '+',  '-',  '+',  '+',  '+', /* C0 */
	 '+',  '+',  '+',  '+',  '+',  '-',  '+',  '+',
	 '+',  '+',  '+',  '+',  '+',  '+',  '+',  '+',
	 '+',  '+',  '+',  ' ',  ' ',  ' ',  ' ',  ' ',
	 'a',  'b',  'G',  'P',  'S',  's',  'u',  'r',
	 'O',  'T',  'O',  'd',  'i',  'f',  'e',  'N',
	 '=',  '=',  '>',  '<',  'f',  'f',  '/',  '=',
	 'o',  '.',  '.',  'R',  'n',  '2',  '.',  ' '
};

static int tab_ibmtolatin1[256] = {
	0xa0,  '0',  '0',  '.',  '.',  '.',  '.',  '.', /* 00 */
	0x08, 0x09, 0x0a,  'o',  'o', 0x0d,  '.', 0xa4,
	 '>',  '<',  '|',  '!', 0xb6, 0xa7,  '_',  '|',
	 '^',  'v',  '>',  '<',  '-',  '-',  '^',  'v',
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, /* 70 */
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e,  '^',
	0xc7, 0xfc, 0xe9, 0xe2, 0xe4, 0xe0, 0xe5, 0xe7, /* 80 */
	0xea, 0xeb, 0xe8, 0xef, 0xee, 0xec, 0xc4, 0xc5,
	0xc9, 0xe6, 0xc6, 0xd4, 0xf6, 0xf2, 0xfb, 0xf9, /* 90 */
	0xff, 0xd6, 0xdc, 0xa2, 0xa3, 0xa5,  'P',  'f',
	0xe1, 0xed, 0xf3, 0xfa, 0xf1, 0xd1, 0xaa, 0xba, /* A0 */
	0xbf, 0xac, 0xac, 0xbd, 0xbc, 0xa1, 0xab, 0xbb,
	 '#',  '#',  '#',  '|',  '+',  '+',  '+',  '+', /* B0 */
	 '+',  '+',  '|',  '+',  '+',  '+',  '+',  '+',
	 '+',  '+',  '+',  '+',  '-',  '+',  '+',  '+', /* C0 */
	 '+',  '+',  '+',  '+',  '+',  '-',  '+',  '+',
	 '+',  '+',  '+',  '+',  '+',  '+',  '+',  '+', /* D0 */
	 '+',  '+',  '+',  ' ',  ' ',  ' ',  ' ',  ' ',
	 'a', 0xdf,  'G',  'P',  'S',  's', 0xb5,  'r', /* E0 */
	 'O',  'T',  'O',  'd',  'i',  'f',  'e',  'N',
	 '=', 0xb1,  '>',  '<',  'f',  'f', 0xf7,  '=', /* F0 */
	0xb0, 0xb7, 0xb7,  'R',  'n', 0xb2, 0xb7, 0xa0
};

#if 0
static int tab_mactoascii[256] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
	 'A',  'A',  'C',  'E',  'N',  'O',  'U',  'a',
	 'a',  'a',  'a',  'a',  'a',  'c',  'e',  'e',
	 'e',  'e',  'i',  'i',  'i',  'i',  'n',  'o',
	 'o',  'o',  'o',  'o',  'u',  'u',  'u',  'u',
	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,
	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
};
#endif


static
void prt_help (void)
{
	fputs (
		"Usage: subst [options] [input [output]]\n"
		"  -i string             Set input file name\n"
		"  -o string             Set output file name\n"
		"  -b, --byte old new    Set a byte substitution\n"
		"  -n, --no-buffer       Process on byte at a time [no]\n"
		"  -s, --string old new  Set a string substitution\n"
		"  -t, --table string    Set a predefined translation table\n"
		"\nTranslation tables are:\n"
		"  lcase  Convert to lower case\n"
		"  ucase  Convert to upper case\n"
		"  nobin  Remove binary characters (0..31, 127..255)\n"
		"  rot13  Encode/decode using rot13\n",
		stdout
	);

	fflush (stdout);
}

static
unsigned subst_read (FILE *fp, unsigned char *buf, unsigned max)
{
	int c;

	if (par_nobuf) {
		c = fgetc (fp);

		if (c == EOF) {
			return (0);
		}

		buf[0] = c;

		return (1);
	}

	return (fread (buf, 1, max, fp));
}

static
int subst_buf (FILE *fp, unsigned char *buf, unsigned n)
{
	unsigned i, j;
	unsigned c;

	if (par_eof) {
		return (0);
	}

	j = 0;

	for (i = 0; i < n; i++) {
		if (par_have_eof) {
			if (buf[i] == par_eof_char) {
				par_eof = 1;
				break;
			}
		}

		c = tab_byte[buf[i]];

		if (c <= 255) {
			if (tab_string[c] != NULL) {
				if (j > 0) {
					if (fwrite (buf, 1, j, fp) != j) {
						return (1);
					}

					j = 0;
				}

				fputs (tab_string[c], fp);
			}
			else {
				buf[j] = (unsigned char) c;
				j += 1;
			}
		}
	}

	if (j > 0) {
		if (fwrite (buf, 1, j, fp) != j) {
			return (1);
		}
	}

	if (par_nobuf) {
		fflush (fp);
	}

	return (0);
}

static
int subst (FILE *out, FILE *inp)
{
	unsigned      n;
	unsigned char buf[SUBST_BUF_SIZE];

	n = subst_read (inp, buf, SUBST_BUF_SIZE);

	while (n > 0) {
		if (subst_buf (out, buf, n)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = subst_read (inp, buf, SUBST_BUF_SIZE);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	return (0);
}

static
unsigned char str_get_hex (char **str)
{
	unsigned      i;
	unsigned char ret;

	ret = 0;

	for (i = 0; i < 2; i++) {
		unsigned char c;

		c = **str;

		if ((c >= '0') && (c <= '9')) {
			ret = 16 * ret + (c - '0');
		}
		else if ((c >= 'a') && (c <= 'f')) {
			ret = 16 * ret + (c - 'a' + 10);
		}
		else if ((c >= 'A') && (c <= 'F')) {
			ret = 16 * ret + (c - 'A' + 10);
		}
		else {
			return (ret);
		}

		*str += 1;
	}

	return (ret);
}

static
void str_expand (char *str)
{
	char *dst;

	dst = str;

	while (1) {
		switch (*str) {
		case 0:
			*dst = 0;
			return;

		case '\\':
			if (str[1] == 'n') {
				*(dst++) = '\n';
				str += 2;
			}
			else if (str[1] == 't') {
				*(dst++) = '\t';
				str += 2;
			}
			else if (str[1] == 'a') {
				*(dst++) = '\a';
				str += 2;
			}
			else if (str[1] == 'b') {
				*(dst++) = '\b';
				str += 2;
			}
			else if (str[1] == '\\') {
				*(dst++) = '\\';
				str += 2;
			}
			else if (str[1] == 'x') {
				str += 2;
				*(dst++) = str_get_hex (&str);
			}
			else {
				*(dst++) = *(str++);
			}
			break;

		default:
			*(dst++) = *(str++);
			break;
		}
	}
}

static
int get_byte_num (const char *str, unsigned *c)
{
	unsigned base;
	unsigned digit;

	base = 10;

	if (str[0] == 0) {
		*c = 256;
		return (0);
	}

	if (str[0] == '0') {
		str += 1;

		if ((str[0] == 'x') || (str[0] == 'X')) {
			base = 16;
			str += 1;
		}
		else if ((str[0] == 'o') || (str[0] == 'O')) {
			base = 8;
			str += 1;
		}
		else if ((str[0] == 'b') || (str[0] == 'B')) {
			base = 2;
			str += 1;
		}
	}

	*c = 0;

	while (1) {
		if ((*str >= '0') && (*str <= '9')) {
			digit = (unsigned) (*str - '0');
		}
		else if ((*str >= 'a') && (*str <= 'z')) {
			digit = (unsigned) (*str - 'a' + 10);
		}
		else if ((*str >= 'A') && (*str <= 'Z')) {
			digit = (unsigned) (*str - 'Z' + 10);
		}
		else if (*str == 0) {
			return (0);
		}
		else {
			return (1);
		}

		if (digit >= base) {
			return (1);
		}

		*c = base * *c + digit;

		str += 1;
	}

	return (1);
}

static
int get_byte (const char *str, unsigned *c)
{
	if ((str[0] >= '0') && (str[0] <= '9')) {
		return (get_byte_num (str, c));
	}

	if ((strcmp (str, "null") == 0) || (strcmp (str, "-1") == 0)) {
		*c = 256;
		return (0);
	}

	*c = (unsigned char) str[0];

	return (0);
}

static
void set_tab (unsigned i, unsigned n, int *tab)
{
	unsigned j;

	for (j = 0; j < n; j++) {
		tab_byte[i + j] = tab[j];
	}
}

static
void set_tab_clear (void)
{
	unsigned i;

	for (i = 0; i < 256; i++) {
		tab_byte[i] = 256;
	}
}

static
void set_tab_null (void)
{
	unsigned i;

	for (i = 0; i < 256; i++) {
		tab_byte[i] = (int) i;
	}
}

static
void set_tab_lcase (void)
{
	set_tab_null();
	set_tab (64, 32, tab_lcase);
}

static
void set_tab_ucase (void)
{
	set_tab_null();
	set_tab (96, 32, tab_ucase);
}

static
void set_tab_nobin (void)
{
	set_tab_clear();
	set_tab (0, 128, tab_nobin);
}

static
void set_tab_ibmtoascii (void)
{
	set_tab (0, 256, tab_ibmtoascii);
}

static
void set_tab_ibmtolatin1 (void)
{
	set_tab (0, 256, tab_ibmtolatin1);
}

static
void set_tab_rot13 (void)
{
	unsigned i;

	set_tab_null();

	for (i = 0; i < 26; i++) {
		tab_byte['A' + i] = 'A' + (i + 13) % 26;
		tab_byte['a' + i] = 'a' + (i + 13) % 26;
	}
}

int main (int argc, char *argv[])
{
	int      i, r;
	unsigned src, dst;
	const char *finp, *fout;
	FILE       *inp, *out;
	unsigned   fcnt;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (0);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("subst");
			return (0);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	set_tab_null();

	for (i = 0; i < 256; i++) {
		tab_string[i] = NULL;
	}

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing output file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "t", "table")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing tanslation table\n",
					argv0
				);
				return (1);
			}

			if (strcmp (argv[i], "lcase") == 0) {
				set_tab_lcase();
			}
			else if (strcmp (argv[i], "ucase") == 0) {
				set_tab_ucase();
			}
			else if (strcmp (argv[i], "nobin") == 0) {
				set_tab_nobin();
			}
			else if (strcmp (argv[i], "ibmtoascii") == 0) {
				set_tab_ibmtoascii();
			}
			else if (strcmp (argv[i], "ibmtolatin1") == 0) {
				set_tab_ibmtolatin1();
			}
			else if (strcmp (argv[i], "rot13") == 0) {
				set_tab_rot13();
			}
			else {
				fprintf (stderr,
					"%s: unknown translation table (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "b", "byte")) {
			if ((i + 3) > argc) {
				fprintf (stderr,
					"%s: missing byte substitution\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (get_byte (argv[i + 1], &src)) {
				fprintf (stderr, "%s: bad source value (%s)\n",
					argv0, argv[i + 1]
				);
				return (EXIT_FAILURE);
			}

			if (src > 255) {
				fprintf (stderr,
					"%s: source value out of range (%u)\n",
					argv0, src
				);
				return (EXIT_FAILURE);
			}

			if (get_byte (argv[i + 2], &dst)) {
				fprintf (stderr,
					"%s: bad destination value (%s)\n",
					argv0, argv[i + 2]
				);
				return (EXIT_FAILURE);
			}

			if (dst > 256) {
				dst = 256;
			}

			tab_byte[src] = dst;

			i += 2;
		}
		else if (str_isarg2 (argv[i], "e", "end-of-file")) {
			if ((i + 1) > argc) {
				fprintf (stderr,
					"%s: missing end of file byte\n", argv0
				);
				return (EXIT_FAILURE);
			}

			if (get_byte (argv[i + 1], &par_eof_char)) {
				fprintf (stderr,
					"%s: bad end of file value (%s)\n",
					argv0, argv[i + 1]
				);
				return (EXIT_FAILURE);
			}

			par_have_eof = 1;

			i += 1;
		}
		else if (str_isarg2 (argv[i], "n", "no-buffer")) {
			par_nobuf = 1;
		}
		else if (str_isarg2 (argv[i], "s", "string")) {
			if ((i + 3) > argc) {
				fprintf (stderr,
					"%s: missing string substitution\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (get_byte (argv[i + 1], &src)) {
				fprintf (stderr,
					"%s: bad source value (%s)\n",
					argv0, argv[i + 1]
				);
				return (EXIT_FAILURE);
			}

			if (src > 255) {
				fprintf (stderr,
					"%s: source value out of range (%u)\n",
					argv0, src
				);
				return (EXIT_FAILURE);
			}

			free (tab_string[src]);
			tab_string[src] = strdup (argv[i + 2]);
			str_expand (tab_string[src]);

			i += 2;
		}
		else if (str_isarg2 (argv[i], NULL, "lcase")) {
			set_tab_lcase();
		}
		else if (str_isarg2 (argv[i], NULL, "ucase")) {
			set_tab_ucase();
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	inp = (finp == NULL) ? stdin : fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			argv0, finp
		);
		return (EXIT_FAILURE);
	}

	out = (fout == NULL) ? stdout : fopen (fout, "wb");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			argv0, fout
		);
		return (EXIT_FAILURE);
	}

	r = subst (out, inp);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
